import abjad
import auxjad


def test_simplified_time_signature_ratio_01():
    time_signature = auxjad.simplified_time_signature_ratio((4, 8))
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((2, 4))')


def test_simplified_time_signature_ratio_02():
    time_signature = auxjad.simplified_time_signature_ratio((7, 8))
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((7, 8))')


def test_simplified_time_signature_ratio_03():
    time_signature = auxjad.simplified_time_signature_ratio((4, 8),
                                                            min_denominator=8,
                                                            )
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((4, 8))')


def test_simplified_time_signature_ratio_04():
    time_signature = auxjad.simplified_time_signature_ratio((4, 8),
                                                            min_denominator=2,
                                                            )
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((1, 2))')


def test_simplified_time_signature_ratio_05():
    time_signature = auxjad.simplified_time_signature_ratio((1, 1))
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((4, 4))')


def test_simplified_time_signature_ratio_06():
    time_signature = auxjad.simplified_time_signature_ratio((1, 1),
                                                            min_denominator=1,
                                                            )
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((1, 1))')


def test_simplified_time_signature_ratio_01():
    arg = (4, 8)
    time_signature = auxjad.simplified_time_signature_ratio(arg)
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((2, 4))')

    arg = abjad.Duration((4, 8))
    time_signature = auxjad.simplified_time_signature_ratio(arg)
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((2, 4))')

    arg = abjad.Meter((4, 8))
    time_signature = auxjad.simplified_time_signature_ratio(arg)
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((2, 4))')

    arg = abjad.TimeSignature((4, 8))
    time_signature = auxjad.simplified_time_signature_ratio(arg)
    assert format(time_signature) == abjad.String.normalize(
        r'abjad.TimeSignature((2, 4))')

    arg = (4, 8)
    pair = auxjad.simplified_time_signature_ratio(arg,
                                                  output_pair_of_int=True,
                                                  )
    assert pair == (2, 4)

    arg = abjad.Duration((4, 8))
    pair = auxjad.simplified_time_signature_ratio(arg,
                                                  output_pair_of_int=True,
                                                  )
    assert pair == (2, 4)

    arg = abjad.Meter((4, 8))
    pair = auxjad.simplified_time_signature_ratio(arg,
                                                  output_pair_of_int=True,
                                                  )
    assert pair == (2, 4)

    arg = abjad.TimeSignature((4, 8))
    pair = auxjad.simplified_time_signature_ratio(arg,
                                                  output_pair_of_int=True,
                                                  )
    assert pair == (2, 4)



from .CartographySelector import CartographySelector
from .LeafDynMaker import LeafDynMaker
from .LoopByList import LoopByList
from .LoopByNotes import LoopByNotes
from .LoopByWindow import LoopByWindow
from .Shuffler import Shuffler
from .TenneySelector import TenneySelector



import copy
import random
import abjad


class AdriftFadeIn():
    r"""xxx
    """

    ### INITIALISER ###

    def __init__(self,
                 container: abjad.Container,
                 ):
        super().__init__(container)

    ### PRIVATE METHODS ###

    def _fade_process(self):
        r"""Custom fade in process.
        """
        pass

    def _create_empty_container(self):
        r'xxx'
        self._current_container = abjad.Container()
        duration = abjad.inspect(self._initial_container).duration()
        multimeasure_rest = abjad.MultimeasureRest(
            (4, 4),
            multiplier=abjad.Multiplier(duration),
        )
        abjad.attach(abjad.TimeSignature(duration), multimeasure_rest)
        self._current_container.append(multimeasure_rest)



import abjad
import pytest
import auxjad


def test_respell_chord_01():
    chord = abjad.Chord("<c' cs'>4")
    auxjad.respell_chord(chord)
    assert format(chord) == abjad.String.normalize(
        r"""
        <c' df'>4
        """)


def test_respell_chord_02():
    staff = abjad.Staff()
    for pitch in range(12):
        chord = abjad.Chord([pitch, pitch + 1], (1, 16))
        auxjad.respell_chord(chord)
        staff.append(chord)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            <c' df'>16
            <cs' d'>16
            <d' ef'>16
            <ds' e'>16
            <e' f'>16
            <f' gf'>16
            <fs' g'>16
            <g' af'>16
            <gs' a'>16
            <a' bf'>16
            <as' b'>16
            <b' c''>16
        }
        """)


def test_respell_chord_03():
    chord = abjad.Chord(r"<a c' cs' f'>1")
    auxjad.respell_chord(chord)
    assert format(chord) == abjad.String.normalize(
        r"""
        <a c' df' f'>1
        """)


def test_respell_chord_04():
    chord = abjad.Chord(r"<e' cs' g' ef'>1")
    auxjad.respell_chord(chord)
    assert format(chord) == abjad.String.normalize(
        r"""
        <cs' ds' e' g'>1
        """)


def test_respell_chord_05():
    chord = abjad.Chord(r"<c' cs''>1")
    auxjad.respell_chord(chord)
    assert format(chord) == abjad.String.normalize(
        r"""
        <c' cs''>1
        """)
    auxjad.respell_chord(chord, include_multiples=True)
    assert format(chord) == abjad.String.normalize(
        r"""
        <c' df''>1
        """)


def test_respell_chord_06():
    chord = abjad.Chord(r"<c' cs' cs''>1")
    auxjad.respell_chord(chord)
    assert format(chord) == abjad.String.normalize(
        r"""
        <c' df' cs''>1
        """)
    chord = abjad.Chord(r"<c' cs' cs''>1")
    auxjad.respell_chord(chord, respell_by_pitch_class=True)
    assert format(chord) == abjad.String.normalize(
        r"""
        <c' df' df''>1
        """)



import abjad
from .simplified_time_signature_ratio import simplified_time_signature_ratio


def rests_to_multimeasure_rest(container: abjad.Container):
    r"""Mutates an input container (of type ``abjad.Container`` or child class)
    in place and has no return value. This function looks for bars filled with
    regular rests and converts them into an ``abjad.MultimeasureRest``.

    ..  container:: example

        Converts any measure filled with regular rests into a measure with a
        single multi-measure rest.

        >>> container = abjad.Container(r"r1")
        >>> auxjad.rests_to_multimeasure_rest(container)
        >>> abjad.f(container)
        {
            R1
        }

        .. figure:: ../_images/image-rests_to_multimeasure_rest-1.png

    ..  container:: example

        Works with measures with multiple regular rests.

        >>> container = abjad.Container(r"\time 3/4 r4 r8.. r32 r4")
        >>> auxjad.rests_to_multimeasure_rest(container)
        >>> abjad.f(container)
        {
            %%% \time 3/4 %%%
            R1 * 3/4
        }

        .. figure:: ../_images/image-rests_to_multimeasure_rest-2.png

    .. note::

        Notice that the time signatures in the output are commented out with
        ``%%%``. This is because Abjad only applies time signatures to
        containers that belong to a ``abjad.Staff``. The present function works
        with either ``abjad.Container`` and ``abjad.Staff``.

        >>> container = abjad.Container(r"\time 3/4 r4 r8.. r32 r4")
        >>> auxjad.rests_to_multimeasure_rest(container)
        >>> abjad.f(container)
        {
            %%% \time 3/4 %%%
            R1 * 3/4
        }
        >>> staff = abjad.Staff([container])
        >>> abjad.f(container)
        {
            %%% \time 3/4 %%%
            R1 * 3/4
        }

    ..  container:: example

        Works with containers with multiple time signatures as well as notes.

        >>> container = abjad.Container(r"\time 3/4 r2. | "
        ...                              "\time 6/8 r2. | "
        ...                              "\time 5/4 c'1 ~ c'4 | r1 r4"
        ...                              )
        >>> auxjad.rests_to_multimeasure_rest(container)
        >>> abjad.f(container)
        {
            %%% \time 3/4 %%%
            R1 * 3/4
            %%% \time 6/8 %%%
            R1 * 3/4
            %%% \time 5/4 %%%
            c'1
            ~
            c'4
            R1 * 5/4
        }

        .. figure:: ../_images/image-rests_to_multimeasure_rest-5.png
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("argument must be 'abjad.Container' or child class")
    leaves = abjad.select(container).leaves()
    for measure in leaves.group_by_measure():
        if all([isinstance(leaf, abjad.Rest) for leaf in measure]):
            first_leaf = abjad.select(measure).leaf(0)
            time_signature = abjad.inspect(first_leaf).indicator(
                abjad.TimeSignature)
            duration = abjad.inspect(measure).duration()
            if duration == 1:
                multiplier = None
            else:
                multiplier = abjad.Multiplier(duration)
            multimeasure_rest = abjad.MultimeasureRest((4, 4),
                                                       multiplier=multiplier,
                                                       )
            if time_signature:
                abjad.attach(time_signature, multimeasure_rest)
            abjad.mutate(measure).replace(multimeasure_rest)



import abjad


def simplified_time_signature_ratio(ratio: (tuple,
                                            abjad.TimeSignature,
                                            abjad.Duration,
                                            abjad.Meter,
                                            ),
                                    *,
                                    min_denominator: int = 4,
                                    output_pair_of_int: bool = False,
                                    ) -> (abjad.TimeSignature, tuple):
    r"""Returns an ``abjad.TimeSignature`` with the simplified ratio of an
    input ratio according to a minimum denominator value. The input ratio can
    be a ``tuple`` of integers, an ``abjad.TimeSignature``, ``abjad.Duration``,
    or an ``abjad.Meter``).

    ..  container:: example

        By default, the function simplifies the ratio of numerator/denominator
        using a minimum denominator value of 4 (that is, the denominator will
        not get smaller than 4). In the case below, (2, 4) is the simplest
        representation of the ratio (4, 8) with a denominator equal to or
        larger than 4.

        >>> time_signature = auxjad.simplified_time_signature_ratio((4, 8))
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))
        >>> time_signature = auxjad.simplified_time_signature_ratio((1, 1))
        >>> format(time_signature)
        abjad.TimeSignature((4, 4))

        If a ratio cannot be simplified at all, the function returns a time
        signature with the original ratio.

        >>> time_signature = auxjad.simplified_time_signature_ratio((7, 8))
        >>> format(time_signature)
        abjad.TimeSignature((7, 8))

        The ``min_denominator`` can be set to values other than 4. If set to 2,
        the simplest representaion of the ratio (4, 8) becomes (1, 2).

        >>> time_signature = auxjad.simplified_time_signature_ratio(
        ...     (4, 8),
        ...     min_denominator=2,
        ... )
        >>> format(time_signature)
        abjad.TimeSignature((1, 2))
        >>> time_signature = auxjad.simplified_time_signature_ratio(
        ...     (1, 1),
        ...     min_denominator=1,
        ... )
        >>> format(time_signature)
        abjad.TimeSignature((1, 1))

    ..  container:: example

        By default, the function returns an ``abjad.TimeSignature`` for
        whatever type of argument it receives (which can be a ``tuple`` of
        integers, an ``abjad.TimeSignature``, an ``abjad.Duration``, or an
        ``abjad.Meter``).

        >>> arg = (4, 8)
        >>> time_signature = auxjad.simplified_time_signature_ratio(arg)
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))

        >>> arg = abjad.Duration((4, 8))
        >>> time_signature = auxjad.simplified_time_signature_ratio(arg)
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))

        >>> arg = abjad.Meter((4, 8))
        >>> time_signature = auxjad.simplified_time_signature_ratio(arg)
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))

        >>> arg = abjad.TimeSignature((4, 8))
        >>> time_signature = auxjad.simplified_time_signature_ratio(arg)
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))

        Call the function with the keyword argument ``output_pair_of_int`` set
        to ``True`` and the output will be a ``tuple`` of integers, regardless
        of the input argument.

        >>> arg = (4, 8)
        >>> pair = auxjad.simplified_time_signature_ratio(
        ...     arg,
        ...     output_pair_of_int=True,
        ... )
        >>> pair
        (2, 4)

        >>> arg = abjad.Duration((4, 8))
        >>> pair = auxjad.simplified_time_signature_ratio(
        ...     arg,
        ...     output_pair_of_int=True,
        ... )
        >>> assert pair == (2, 4)

        arg = abjad.Meter((4, 8))
        >>> pair = auxjad.simplified_time_signature_ratio(
        ...     arg,
        ...     output_pair_of_int=True,
        ... )
        >>> pair
        (2, 4)

        arg = abjad.TimeSignature((4, 8))
        >>> pair = auxjad.simplified_time_signature_ratio(
        ...     arg,
        ...     output_pair_of_int=True,
        ... )
        >>> pair
        (2, 4)
    """
    if not isinstance(ratio, (tuple, abjad.TimeSignature,
                              abjad.Duration, abjad.Meter)):
        raise TypeError("argument must be 'tuple', 'abjad.Duration', or "
                        "'abjad.TimeSignature'")
    if not isinstance(min_denominator, int):
        raise TypeError("'min_denominator' must be 'int'")
    if not isinstance(output_pair_of_int, bool):
        raise TypeError("'output_pair_of_int' must be 'bool'")

    if isinstance(ratio, (abjad.TimeSignature, abjad.Duration, abjad.Meter)):
        numerator, denominator = ratio.pair
    else:
        numerator, denominator = ratio
        if not isinstance(numerator, int):
            raise TypeError("'ratio's elements must be 'int'")
        if not isinstance(denominator, int):
            raise TypeError("'ratio's elements must be 'int'")

    while denominator < min_denominator:
        numerator *= 2
        denominator *= 2
    while True:
        if (numerator % 2 == 0 and denominator % 2 == 0
                and denominator >= 2 * min_denominator):
            numerator /= 2
            denominator /= 2
        else:
            if not output_pair_of_int:
                return abjad.TimeSignature((int(numerator), int(denominator)))
            else:
                return int(numerator), int(denominator)



from .close_container import close_container
from .container_is_full import container_is_full
from .containers_are_equal import containers_are_equal
from .fill_with_rests import fill_with_rests
from .leaves_are_tieable import leaves_are_tieable
from .remove_repeated_dynamics import remove_repeated_dynamics
from .remove_repeated_time_signatures import remove_repeated_time_signatures
from .repeat_container import repeat_container
from .respell_container import respell_container
from .respell_chord import respell_chord
from .rests_to_multimeasure_rest import rests_to_multimeasure_rest
from .simplified_time_signature_ratio import simplified_time_signature_ratio
from .sync_containers import sync_containers
from .underfull_duration import underfull_duration



import copy
from math import ceil
import abjad
from ._LoopParent import _LoopParent


class LoopByWindow(_LoopParent):
    r"""This class can be used to output slices of an ``abjad.Container`` using
    the metaphor of a looping window of a constant size given by an
    ``abjad.Duration``.

    ..  container:: example

        Usage is similar to other factory classes. It takes a container (or
        child class equivalent) as argument. Each call of the object, in this
        case ``looper()``, will move the window forwards and output the sliced
        window. If no window_size nor step_size are entered as arguments, they
        are set to the following default values, respectively: (4, 4), i.e. a
        window of the size of a 4/4 bar, and (1, 16), i.e. a step of the length
        of a sixteenth-note.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByWindow-1.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

        .. figure:: ../_images/image-LoopByWindow-2.png

        The property ``current_window`` can be used to access the current
        window without moving the head forwards.

        >>> notes = looper.current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

        .. figure:: ../_images/image-LoopByWindow-3.png

    ..  container:: example

        The very first call will output the input container without processing
        it. To disable this behaviour and have the looping window move on the
        very first call, initialise the class with the keyword argument
        ``move_window_on_first_call`` set to ``True``.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              move_window_on_first_call=True,
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

        .. figure:: ../_images/image-LoopByWindow-4.png

    ..  container:: example

        The optional arguments ``window_size`` and ``step_size`` can be used to
        set different window and step sizes. ``window_size`` can take a tuple
        or an ``abjad.Meter`` as input, while ``step_size`` takes a tuple or an
        ``abjad.Duration``.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }

        .. figure:: ../_images/image-LoopByWindow-5.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByWindow-6.png

    ..  container:: example

        The instances of ``LoopByWindow`` can also be used as an iterator,
        which can then be used in a for loop to exhaust all windows. Notice how
        it appends rests at the end of the container, until it is totally
        exhausted.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 8),
        ...                              )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
            c'8
            d'8
            ~
            d'4.
            e'8
            d'2
            e'4
            d'4.
            e'8
            ~
            e'8
            r8
            d'4
            e'4
            r4
            d'8
            e'8
            ~
            e'8
            r4.
            e'4
            r2
            e'8
            r8
            r2
        }

        .. figure:: ../_images/image-LoopByWindow-7.png

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation, besides ``window_size`` and ``step_size``. ``max_steps`` sets
        the maximum number of steps that the window can advance when the object
        is called, ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition).
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window. It must be a tuple or an
        ``abjad.Duration``, and its default value is 0.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(5, 8),
        ...                              max_steps=2,
        ...                              repetition_chance=0.25,
        ...                              forward_bias=0.2,
        ...                              head_position=(2, 8),
        ...                              omit_time_signature=False,
        ...                              )
        >>> looper.window_size
        3/4
        >>> looper.step_size
        5/8
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        1/4
        >>> looper.omit_time_signature
        False

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = (5, 4)
        >>> looper.step_size = (1, 4)
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 0
        >>> looper.omit_time_signature = True
        >>> looper.window_size
        5/4
        >>> looper.step_size
        1/4
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        0
        >>> looper.omit_time_signature
        True

    ..  container:: example

        The function ``len()`` can be used to get the total number of steps
        in the contents (always rounded up).

        >>> input_music = abjad.Container(r"c'1")
        >>> looper = auxjad.LoopByWindow(input_music)
        >>> len(looper)
        16
        >>> input_music = abjad.Container(r"c'1")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              step_size=(1, 4),
        ...                              )
        >>> len(looper)
        4
        >>> input_music = abjad.Container(r"c'2..")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              step_size=(1, 4),
        ...                              window_size=(2, 4),
        ...                              )
        >>> len(looper)
        4

    ..  container:: example

        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method ``output_all()``.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> music = looper.output_all()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
            e'4
            f'4
            r4
            f'4
            r2
        }

        .. figure:: ../_images/image-LoopByWindow-8.png

    ..  container:: example

        When using ``output_all()``, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> input_music = abjad.Container(r"c'4 <e' f' g'>2 r4 f'2.")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> music = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            <e' f' g'>2
            ~
            <e' f' g'>2
            r4
            <e' f' g'>4
            r4
            f'4
            r4
            f'2
            ~
            f'2.
            ~
            f'2
            r4
            f'4
            r2
        }

        .. figure:: ../_images/image-LoopByWindow-9.png

    ..  container:: example

        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument. Similarly
        to ``output_all()``, the keyword argument ``tie_identical_pitches`` is
        available for tying pitches.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> music = looper.output_n(2)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
        }

        .. figure:: ../_images/image-LoopByWindow-10.png

    .. container:: example

        To change the size of the looping window after instantiation, use the
        property ``window_size``. In the example below, the initial window is
        of size (4, 4), but changes to (3, 8) after three calls. Notice how the
        very first call attaches a time signature equivalent to the window size
        to the output window; subsequent calls will not have time signatures
        unless the size of the looping window changes.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music)
        >>> staff = abjad.Staff()
        >>> for _ in range(3):
        ...     notes = looper()
        ...     staff.append(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
            c'8
            d'8
            ~
            d'4.
            e'8
            ~
            e'8
            f'8
        }

        .. figure:: ../_images/image-LoopByWindow-11.png

        >>> looper.window_size = (3, 8)
        >>> staff = abjad.Staff()
        >>> for _ in range(3):
        ...     notes = looper()
        ...     staff.append(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'16
            d'16
            ~
            d'4
            d'4.
            d'4.
        }

        .. figure:: ../_images/image-LoopByWindow-12.png

        To disable time signatures altogether, initialise ``LoopByWindow`` with
        the keyword argument ``omit_time_signature`` set to ``True`` (default
        is ``False``), or use the ``omit_time_signature`` property after
        initialisation.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music, omit_time_signature=True)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByWindow-13.png

    ..  container:: example

        This class can handle dynamics and articulations too. When a leaf is
        shortened by the looping window's movement, the dynamics and
        articulations are still applied to it.

        >>> input_music = abjad.Container(
        ... r"c'4-.\p\< d'2--\f e'4->\ppp f'2 ~ f'8")
        >>> looper = auxjad.LoopByWindow(input_music)
        >>> staff = abjad.Staff()
        >>> for _ in range(2):
        ...     music = looper()
        ...     staff.append(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            \p
            - \staccato
            \<
            d'2
            \f
            - \tenuto
            e'4
            \ppp
            - \accent
            c'8.
            \p
            - \staccato
            \<
            d'16
            \f
            - \tenuto
            ~
            d'4..
            e'16
            \ppp
            - \accent
            ~
            e'8.
            f'16
        }

        .. figure:: ../_images/image-LoopByWindow-14.png

    .. container:: example

        Use the ``contents`` property to read as well as overwrite the contents
        of the looper. Notice that the ``head_position`` will remain on its
        previous value and must be reset to ``0`` if that's required.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByWindow(input_music)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByWindow-15.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

        .. figure:: ../_images/image-LoopByWindow-16.png

        >>> looper.contents = abjad.Container(r"c'16 d'16 e'16 f'16 g'2. a'1")
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            e'16
            f'16
            g'8
            ~
            g'2
            ~
            g'8
            a'8
        }

        .. figure:: ../_images/image-LoopByWindow-17.png

        >>> looper.head_position = 0
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'16
            d'16
            e'16
            f'16
            g'2.
        }

        .. figure:: ../_images/image-LoopByWindow-18.png

    ..  container:: example

        This class can handle tuplets, but this functionality should be
        considered experimental.

        >>> input_music = abjad.Container(r"\times 2/3 {c'8 d'8 e'} d'2.")
        >>> looper = auxjad.LoopByWindow(input_music,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16))
        >>> staff = abjad.Staff()
        >>> for _ in range(3):
        ...     window = looper()
        ...     staff.append(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \times 2/3 {
                \time 3/4
                c'8
                d'8
                e'8
            }
            d'2
            \times 2/3 {
                c'32
                d'16
                ~
                d'16
                e'8
            }
            d'16
            ~
            d'2
            \times 2/3 {
                d'16
                e'8
            }
            d'8
            ~
            d'2
        }

        .. figure:: ../_images/image-LoopByWindow-19.png
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: abjad.Container,
                 *,
                 window_size: (tuple, abjad.Meter) = (4, 4),
                 step_size: (int, float, tuple, str, abjad.Duration) = (1, 16),
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: (int, float, tuple, str, abjad.Duration) = 0,
                 omit_time_signature: bool = False,
                 move_window_on_first_call: bool = False,
                 ):
        self.contents = contents
        self._new_time_signature = True
        self.omit_time_signature = omit_time_signature
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         move_window_on_first_call,
                         )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(abjad.f(self._contents))

    def __len__(self) -> int:
        r'Outputs the length of ``contents`` in terms of ``step_size``.'
        return ceil(abjad.inspect(self._contents).duration() / self._step_size)

    ### PRIVATE METHODS ###

    def _slice_contents(self):
        r"""This method takes a slice of size ``window_size`` out of the
        contents starting at the current ``head_position``.
        """
        head = self._head_position
        window_size = self._window_size
        dummy_container = copy.deepcopy(self._contents)
        # splitting leaves at both slicing points
        if head > abjad.Duration(0):
            abjad.mutate(dummy_container[:]).split([head,
                                                    window_size.duration,
                                                    ])
        else:
            abjad.mutate(dummy_container[:]).split([window_size.duration])
        # finding start and end indeces for the window
        for start in range(len(dummy_container)):
            if abjad.inspect(dummy_container[:start + 1]).duration() > head:
                break
        for end in range(start + 1, len(dummy_container)):
            if (abjad.inspect(dummy_container[start : end]).duration()
                    == window_size.duration):
                break
        else:
            end = len(dummy_container)
        # passing on indicators from the head of an initial splitted leaf
        for index in range(start - 1, -1, -1):
            if abjad.inspect(dummy_container[index]).indicator(abjad.Tie):
                inspect_contents = abjad.inspect(dummy_container[index - 1])
                if index == 0 or not inspect_contents.indicator(abjad.Tie):
                    inspect_contents = abjad.inspect(dummy_container[index])
                    for indicator in inspect_contents.indicators():
                        if not isinstance(indicator,
                                          (abjad.TimeSignature, abjad.Tie),
                                          ):
                            abjad.attach(indicator, dummy_container[start])
        # removing ties generated by the split mutation
        abjad.detach(abjad.Tie(), dummy_container[start - 1])
        abjad.detach(abjad.Tie(), dummy_container[end - 1])
        # appending rests if necessary
        contents_dur = abjad.inspect(dummy_container[start : end]).duration()
        if contents_dur < window_size.duration:
            missing_dur = window_size.duration - contents_dur
            rests = abjad.LeafMaker()(None, missing_dur)
            dummy_container.extend(rests)
            end += len(rests)
        # transforming abjad.Selection -> abjad.Container for rewrite_meter
        dummy_container = abjad.Container(
            abjad.mutate(dummy_container[start : end]).copy()
        )
        abjad.mutate(dummy_container[:]).rewrite_meter(window_size)
        if self._new_time_signature and not self._omit_time_signature:
            abjad.attach(abjad.TimeSignature(window_size),
                         abjad.select(dummy_container).leaves()[0],
                         )
            self._new_time_signature = False
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self):
        r'The ``list`` which serves as the basis for the slices of the looper.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: abjad.Container,
                 ):
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or "
                            "child class")
        self._contents = copy.deepcopy(contents)
        self._remove_all_time_signatures(self._contents)
        self._contents_length = abjad.inspect(contents[:]).duration()

    @property
    def head_position(self) -> abjad.Duration:
        r'The position of the head at the start of a looping window.'
        return self._head_position

    @head_position.setter
    def head_position(self,
                      head_position: (tuple, abjad.Duration),
                      ):
        r"""This setter method replaces the paren'ts one since the parent's 
        method uses integers as input intead of tuples or ``abjad.Duration``.
        """
        if not isinstance(head_position,
                          (int, float, tuple, str, abjad.Duration),
                          ):
            raise TypeError("'head_position' must be a number, 'tuple', or "
                            "'abjad.Duration'")
        if abjad.Duration(head_position) >= self._contents_length:
            raise ValueError("'head_position' must be smaller than the "
                             "length of 'contents'")
        self._is_first_window = True
        self._head_position = abjad.Duration(head_position)

    @property
    def window_size(self) -> abjad.Meter:
        r'The length of the looping window.'
        return self._window_size

    @window_size.setter
    def window_size(self,
                    window_size: (int, float, tuple, abjad.Meter),
                    ):
        r"""This setter method replaces the paren'ts one since the parent's
        method uses integers as input intead of tuples or ``abjad.Duration``.
        """
        if not isinstance(window_size,
                          (int, float, tuple, str, abjad.Meter),
                          ):
            raise TypeError("'window_size' must be 'tuple' or 'abjad.Meter'")
        if (abjad.Meter(window_size).duration
                > self._contents_length - self._head_position):
            raise ValueError("'window_size' must be smaller than or equal "
                             "to the length of 'contents'")
        if (self._is_first_window or self._window_size.duration
                != abjad.Meter(window_size).duration):
            self._window_size = abjad.Meter(window_size)
            self._new_time_signature = True

    @property
    def step_size(self) -> abjad.Duration:
        r'The size of each step when moving the head.'
        return self._step_size

    @step_size.setter
    def step_size(self,
                  step_size: (tuple, abjad.Duration),
                  ):
        r"""This setter method replaces the paren'ts one since the parent's
        method uses integers as input intead of tuples or ``abjad.Duration``.
        """
        if not isinstance(step_size,
                          (int, float, tuple, str, abjad.Duration),
                          ):
            raise TypeError("'step_size' must be a 'tuple' or "
                            "'abjad.Duration'")
        self._step_size = abjad.Duration(step_size)

    @property
    def omit_time_signature(self) -> list:
        r'When ``True``, the output will contain no time signatures.'
        return self._omit_time_signature

    @omit_time_signature.setter
    def omit_time_signature(self,
                            omit_time_signature: bool,
                            ):
        if not isinstance(omit_time_signature, bool):
            raise TypeError("'omit_time_signature' must be 'bool'")
        self._omit_time_signature = omit_time_signature

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r"""Boolean indicating whether the process is done (i.e. whether the
        head position has overtaken the ``contents`` length).

        This property replaces the parent's one since the parent's property
        uses the number of indeces of ``contents``.
        """
        return (self._head_position >= self._contents_length
            or self._head_position < 0)



import auxjad
import abjad
import random


def test_complex_music_example_02():
    random.seed(87435)
    container = abjad.Staff([
        auxjad.ArtificialHarmonic("<ds' gs'>4"),
        auxjad.ArtificialHarmonic("<b ds'>8."),
        auxjad.ArtificialHarmonic("<g c'>2.", is_parenthesized=True),
        abjad.Rest("r4"),
        abjad.Chord([0, 1, 7], (1, 8)),
        auxjad.ArtificialHarmonic("<d' a'>2.", is_parenthesized=True),
    ])
    # adding a time signature to the first note
    container_length = abjad.inspect(container).duration()
    abjad.attach(abjad.TimeSignature(container_length), container[0])
    # respelling augmented unisons
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 37/16
            <
                ds'
                \tweak style #'harmonic
                gs'
            >4
            <
                b
                \tweak style #'harmonic
                ds'
            >8.
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2.
            r4
            <c' df' g'>8
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >2.
        }
        """)
    # Using a looping window by elements 3 times
    looper = auxjad.LoopByNotes(container, window_size=4)
    staff = abjad.Staff()
    for _ in range(3):
        music = looper()
        staff.append(music)
    # shuffling the last output container by the looping window 3 times
    container = abjad.Container(looper.current_window)
    shuffler = auxjad.Shuffler(container, omit_time_signatures=True)
    for _ in range(3):
        music = shuffler()
        staff.append(music)
    # adding initial dynamics
    abjad.attach(abjad.Dynamic('ppp'), staff[0])
    assert format(container) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 23/16
            <
                ds'
                \tweak style #'harmonic
                gs'
            >4
            \ppp
            <
                b
                \tweak style #'harmonic
                ds'
            >8.
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2.
            r4
            \time 21/16
            <
                b
                \tweak style #'harmonic
                ds'
            >8.
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2.
            r4
            <c' df' g'>8
            \time 15/8
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2.
            r4
            <c' df' g'>8
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >2.
            r4
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >4.
            ~
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >4.
            <c' df' g'>8
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2.
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >2.
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2
            ~
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >4
            r8
            r8
            <c' df' g'>8
            <c' df' g'>8
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >2
            ~
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                g
                \tweak style #'harmonic
                c'
            >4
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >4.
            ~
            <
                \parenthesize
                \tweak ParenthesesItem.font-size #-4
                d'
                \tweak style #'harmonic
                a'
            >4.
            r4
        }
        """)


def test_complex_music_example_02():
    random.seed(1242)
    # selectors of raw materials
    pitch_selector = auxjad.TenneySelector([0, 7, 8, 2, 3, 10])
    duration_selector = auxjad.CartographySelector([(2, 8), (3, 8), (5, 8)])
    dynamic_selector = auxjad.CartographySelector(['p', 'mp', 'mf', 'f'])
    articulation_selector = auxjad.CartographySelector([None, '-', '>'])
    # creating notes
    pitches = [pitch_selector() for _ in range(8)]
    durations = [duration_selector() for _ in range(8)]
    dynamics = [dynamic_selector() for _ in range(8)]
    articulations = [articulation_selector() for _ in range(8)]
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
    container = abjad.Staff(notes)
    # adding a time signature to the first note
    container_length = abjad.inspect(container).duration()
    abjad.attach(abjad.TimeSignature(container_length), container[0])
    assert format(container) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 13/4
            c'2
            \mp
            - \tenuto
            ~
            c'8
            af'4.
            \mp
            bf'4.
            \mf
            - \tenuto
            c'4.
            \mp
            d'4.
            \mf
            - \accent
            ef'4
            \p
            - \accent
            af'2
            \mp
            ~
            af'8
            c'4
            \mp
            - \accent
        }
        """)
    # Using a looping window 3 times with the container created above as input
    looper = auxjad.LoopByWindow(container)
    staff = abjad.Staff()
    for _ in range(3):
        music = looper()
        staff.append(music)
    # shuffling the last output container by the looping window 3 times
    container = abjad.Container(looper.current_window)
    shuffler = auxjad.Shuffler(container, omit_time_signatures=True)
    for _ in range(3):
        music = shuffler()
        staff.append(music)
    # continuing with the looping process 3 more times using the last shuffled
    # container
    container = abjad.Container(shuffler.current_window)
    looper = auxjad.LoopByWindow(container, window_size=(3, 4))
    for _ in range(3):
        music = looper()
        staff.append(music)
    # removing repeated dynamics
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'2
            \mp
            - \tenuto
            ~
            c'8
            af'4.
            c'2
            - \tenuto
            ~
            c'16
            af'8.
            ~
            af'8.
            bf'16
            \mf
            - \tenuto
            c'2
            \mp
            - \tenuto
            af'4.
            bf'8
            \mf
            - \tenuto
            bf'8
            - \tenuto
            c'8
            \mp
            - \tenuto
            ~
            c'4.
            af'4.
            c'2
            - \tenuto
            bf'8
            \mf
            - \tenuto
            af'4.
            \mp
            bf'8
            \mf
            - \tenuto
            c'8
            \mp
            - \tenuto
            ~
            c'4.
            af'4.
            \time 3/4
            bf'8
            \mf
            - \tenuto
            c'8
            \mp
            - \tenuto
            ~
            c'4.
            af'8
            bf'16
            \mf
            - \tenuto
            c'8.
            \mp
            - \tenuto
            ~
            c'4
            ~
            c'16
            af'8.
            c'2
            - \tenuto
            af'4
        }
        """)



import abjad
from .respell_chord import respell_chord


def respell_container(container: abjad.Container,
                      *,
                      include_multiples: bool = False,
                      respell_by_pitch_class: bool = False,
                      ):
    r"""Mutates an input container (of type ``abjad.Container`` or child class)
    in place and has no return value. This function changes the accidentals of
    individual pitches of all chords in a container in order to avoid augmented
    unisons.

    ..  container:: example

        To use this function, apply it to a container that contains chords
        with augmented unisons.

        >>> container = abjad.Container(r"c'4 r4 <ef' e'>4 g'4 <c' cs'>4 r2.")
        >>> auxjad.respell_container(container)
        >>> abjad.f(container)
        {
            c'4
            r4
            <ds' e'>4
            g'4
            <c' df'>4
            r2.
        }

        .. figure:: ../_images/image-respell_container-1.png

    ..  container:: example

        The example below shows the default spelling of 2-note chords by
        ``Abjad`` in the upper staff, and the respelt 2-note chords in the
        bottom staff.

        >>> staff1 = abjad.Staff()
        >>> staff2 = abjad.Staff()
        >>> for pitch in range(12):
        ...     staff1.append(abjad.Chord([pitch, pitch + 1], (1, 16)))
        ...     staff2.append(abjad.Chord([pitch, pitch + 1], (1, 16)))
        >>> auxjad.respell_container(staff2)
        >>> literal = abjad.LilyPondLiteral(r'\accidentalStyle dodecaphonic')
        >>> abjad.attach(literal, staff1)
        >>> abjad.attach(literal, staff2)
        >>> score = abjad.Score([staff1, staff2])
        >>> abjad.f(score)
        \new Score
        <<
            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' cs'>16
                <cs' d'>16
                <d' ef'>16
                <ef' e'>16
                <e' f'>16
                <f' fs'>16
                <fs' g'>16
                <g' af'>16
                <af' a'>16
                <a' bf'>16
                <bf' b'>16
                <b' c''>16
            }
            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' df'>16
                <cs' d'>16
                <d' ef'>16
                <ds' e'>16
                <e' f'>16
                <f' gf'>16
                <fs' g'>16
                <g' af'>16
                <gs' a'>16
                <a' bf'>16
                <as' b'>16
                <b' c''>16
            }
        >>

        .. figure:: ../_images/image-respell_container-2.png

    ..  container:: example

        The function looks for all augmented unissons in chords of 3 or more
        pitches:

        >>> container1 = abjad.Container(r"<a c' cs' f'>1")
        >>> container2 = abjad.Container(r"<a c' cs' f'>1")
        >>> auxjad.respell_container(container2)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <a c' cs' f'>1
            }
            {
                <a c' df' f'>1
            }
        }

        .. figure:: ../_images/image-respell_container-3.png

        It is not a problem if the pitches are input out of order.

        >>> container1 = abjad.Container(r"<e' cs' g' ef'>1")
        >>> container2 = abjad.Container(r"<e' cs' g' ef'>1")
        >>> auxjad.respell_container(container2)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <cs' ef' e' g'>1
            }
            {
                <cs' ds' e' g'>1
            }
        }

        .. figure:: ../_images/image-respell_container-4.png

    ..  container:: example

        By default, this function only changes spelling for pitches that are
        1 semitone apart.

        >>> container1 = abjad.Container(r"<c' cs''>1")
        >>> container2 = abjad.Container(r"<c' cs''>1")
        >>> auxjad.respell_container(container2)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <c' cs''>1
            }
            {
                <c' cs''>1
            }
        }

        .. figure:: ../_images/image-respell_container-5.png

        To consider pitches in different octaves (thus including augmented
        unisons, augmented octaves, augmented fifteenths, etc.), call this
        function with the keyword argument ``include_multiples`` set to
        ``True``.

        >>> container1 = abjad.Container(r"<c' cs''>1")
        >>> container2 = abjad.Container(r"<c' cs''>1")
        >>> auxjad.respell_container(container2, include_multiples=True)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <c' cs''>1
            }
            {
                <c' df''>1
            }
        }

        .. figure:: ../_images/image-respell_container-6.png

    ..  container:: example

        By default, when this function changes the spelling of a pitch, it does
        not change the spelling of all other pitches with the same pitch-class.

        >>> container1 = abjad.Container(r"<c' cs' cs''>1")
        >>> container2 = abjad.Container(r"<c' cs' cs''>1")
        >>> auxjad.respell_container(container2)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <c' cs' cs''>1
            }
            {
                <c' df' cs''>1
            }
        }

        .. figure:: ../_images/image-respell_container-7.png

        To alter all pitch-classes, call this function with the keyword
        argument ``respell_by_pitch_class`` set to ``True``.

        >>> container1 = abjad.Container(r"<c' cs' cs''>1")
        >>> container2 = abjad.Container(r"<c' cs' cs''>1")
        >>> auxjad.respell_container(container2, respell_by_pitch_class=True)
        >>> staff = abjad.Staff([container1, container2])
        >>> abjad.f(staff)
        \new Staff
        {
            {
                <c' cs' cs''>1
            }
            {
                <c' df' df''>1
            }
        }

        .. figure:: ../_images/image-respell_container-8.png
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("argument must be 'abjad.Container' or child class")
    if not isinstance(respell_by_pitch_class, bool):
        raise TypeError("'respell_by_pitch_class' must be 'bool'")
    if not isinstance(include_multiples, bool):
        raise TypeError("'include_multiples' must be 'bool'")
    for leaf in abjad.select(container).leaves():
        if isinstance(leaf, abjad.Chord):
            respell_chord(leaf,
                          include_multiples=include_multiples,
                          respell_by_pitch_class=respell_by_pitch_class,
                          )



import abjad
from ._HarmonicParent import _HarmonicParent


class ArtificialHarmonic(abjad.Chord, _HarmonicParent):
    r"""Creates an chord with a tweaked top note head for notating artificial
    harmonics. This is a child class of ``abjad.Chord``.

    ..  container:: example

        Usage is similar to ``abjad.Chord``:

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4")
        >>> harm.style
        'harmonic'
        >>> abjad.f(harm)
        <
            g
            \tweak style #'harmonic
            c'
        >4

        .. figure:: ../_images/image-ArtificialHarmonic-1.png

        And similarly to ``abjad.Chord``, pitch and duration can be input in
        many different ways:

        >>> harm1 = auxjad.ArtificialHarmonic("<g c'>4")
        >>> harm2 = auxjad.ArtificialHarmonic(["g", "c'"], 1/4)
        >>> harm3 = auxjad.ArtificialHarmonic([-5, 0], 0.25)
        >>> harm4 = auxjad.ArtificialHarmonic([-5, 0], abjad.Duration(1, 4))
        >>> staff = abjad.Staff([harm1, harm2, harm3, harm4])
        >>> abjad.f(staff)
        \new Staff
        {
            <
                g
                \tweak style #'harmonic
                c'
            >4
            <
                g
                \tweak style #'harmonic
                c'
            >4
            <
                g
                \tweak style #'harmonic
                c'
            >4
            <
                g
                \tweak style #'harmonic
                c'
            >4
        }

        .. figure:: ../_images/image-ArtificialHarmonic-2.png



    ..  container:: example

        It is important to note that this class can only be initialised with
        exactly two pitches. Any other number of pitches will raise a
        ``ValueError``:

        >>> auxjad.ArtificialHarmonic("<g c' d'>4")
        ValueError: 'ArtificialHarmonic' requires exactly two 'note_heads' for
        initialisation

    ..  container:: example

        When creating an ``ArtificialHarmonic``, use the keyword argument
        ``style`` to set a different type of note head for the top note, such
        as ``'harmonic-mixed'``:

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4",
        ...                                  style='harmonic-mixed',
        ...                                  )
        >>> harm.style
        'harmonic-mixed'
        >>> abjad.f(harm)
        <
            g
            \tweak style #'harmonic-mixed
            c'
        >4

        .. figure:: ../_images/image-ArtificialHarmonic-3.png

    ..  container:: example

        To notate natural harmonics with a parenthesised pitch for the open
        string at the bottom of the interval, set the keyword
        ``is_parenthesized`` to ``True``.

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4",
        ...                                  is_parenthesized=True,
        ...                                  )
        >>> harm.is_parenthesized
        True
        >>> abjad.f(harm)
        <
            \parenthesize
            \tweak ParenthesesItem.font-size #-4
            g
            \tweak style #'harmonic
            c'
        >4

        .. figure:: ../_images/image-ArtificialHarmonic-4.png

    ..  container:: example

        Similarly to ``abjad.Chord``, ``ArtificialHarmonic`` can take
        multipliers:

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4",
        ...                                  multiplier=(2, 3),
        ...                                  )
        >>> harm.multiplier
        abjad.Multiplier(2, 3)
        >>> abjad.f(harm)
        <
            g
            \tweak style #'harmonic
            c'
        >4 * 2/3

        .. figure:: ../_images/image-ArtificialHarmonic-5.png

    ..  container:: example

        All properties of ``abjad.Chord`` are also available to be read. This
        class also includes two new properties named ``style`` and
        ``is_parenthesized``:

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4")
        >>> harm.written_pitches
        "g c'"
        >>> harm.written_duration
        1/4
        >>> harm.style
        'harmonic'
        >>> harm.is_parenthesized
        False

        All these properties can be set to different values after
        initialisation:

        >>> harm.written_pitches = [-5, 2]
        >>> harm.written_duration = abjad.Duration(1, 8)
        >>> harm.style = 'harmonic-mixed'
        >>> harm.is_parenthesized = True
        >>> harm.written_pitches
        "g d'"
        >>> harm.written_duration
        1/8
        >>> harm.style
        'harmonic-mixed'
        >>> harm.is_parenthesized
        True

    ..  container:: example

        The methods ``sounding_pitch()`` and ``sounding_note()`` return the
        sounding pitch and sounding note, respectively. Their types are
        ``abjad.Pitch`` and ``abjad.Note``, respectively.

        >>> harmonics = [ArtificialHarmonic("<g b>4"),
        ...              ArtificialHarmonic("<g c'>4"),
        ...              ArtificialHarmonic("<g d'>4"),
        ...              ArtificialHarmonic("<g e'>4"),
        ...              ArtificialHarmonic("<g g'>4"),
        ...              ]
        >>> for harmonic in harmonics:
        ...     print(harmonic.sounding_pitch())
        b''
        g''
        d''
        b''
        g'
        >>> for harmonic in harmonics:
        ...     print(harmonic.sounding_note())
        b''4
        g''4
        d''4
        b''4
        g'4

    ..  container:: example

        To add a markup expression to the harmonic note, use the markup:

        >>> harm1 = auxjad.ArtificialHarmonic("<a d'>1")
        >>> harm2 = auxjad.ArtificialHarmonic("<a d'>1",
        ...                                   markup='I.',
        ...                                   )
        >>> harm3 = auxjad.ArtificialHarmonic("<a d'>1",
        ...                                   markup='I.',
        ...                                   direction=abjad.Down)
        >>> staff = abjad.Staff([harm1, harm2, harm3])
        >>> abjad.f(staff)
        \new Staff
        {
            <
                a
                \tweak style #'harmonic
                d'
            >1
            <
                a
                \tweak style #'harmonic
                d'
            >1
            ^ \markup { I. }
            <
                a
                \tweak style #'harmonic
                d'
            >1
            _ \markup { I. }
        }

        .. figure:: ../_images/image-ArtificialHarmonic-6.png

        Setting ``markup`` to ``None`` will remove the markup from the note.

        >>> harm = auxjad.ArtificialHarmonic("<a d'>1",
        ...                                  markup='I.',
        ...                                  )
        >>> harm.markup = None
        >>> abjad.f(harm)
        <
            a
            \tweak style #'harmonic
            d'
        >1

        .. figure:: ../_images/image-ArtificialHarmonic-7.png

    ..  warning::

        If another markup is attached to the harmonic note, trying to set the
        ``markup`` to ``None`` will raise an Exception:

        >>> harm = auxjad.ArtificialHarmonic("<a d'>1")
        >>> abjad.attach(abjad.Markup('test'), harm)
        >>> harm.markup = 'I.'
        >>> harm.markup = None
        Exception: multiple indicators attached to client.

    ..  container:: example

        The note created by ``sounding_note()`` inherits all indicators from
        the ``ArtificialHarmonic``.

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4-.\pp")
        >>> abjad.f(harm.sounding_note())
        g''4
        \pp
        - \staccato

        .. figure:: ../_images/image-ArtificialHarmonic-8.png

    ..  warning::

        Both ``sounding_pitch()`` and ``sounding_note()`` methods raise a
        ValueError exception when it cannot calculate the sounding pitch for
        the given interval.

        >>> ArtificialHarmonic("<g ef'>4").sounding_pitch()
        ValueError: cannot calculate sounding pitch for given interval
        >>> ArtificialHarmonic("<g ef'>4").sounding_note()
        ValueError: cannot calculate sounding pitch for given interval
    """

    ### INITIALISER ###

    def __init__(self,
                 *arguments,
                 multiplier: abjad.typings.DurationTyping = None,
                 tag: abjad.Tag = None,
                 style: str = 'harmonic',
                 is_parenthesized: bool = False,
                 markup: str = None,
                 direction: (str, abjad.enums.VerticalAlignment) = 'up',
                 ):
        super().__init__(*arguments, multiplier=multiplier, tag=tag)
        if len(self.note_heads) != 2:
            raise ValueError("'ArtificialHarmonic' requires exactly two "
                             "'note_heads' for initialisation")
        self.style = style
        self.is_parenthesized = is_parenthesized
        self._direction = direction
        self.markup = markup

    ### PUBLIC METHODS ###

    def sounding_pitch(self) -> abjad.Pitch:
        r'Returns the sounding pitch of the harmonic as an ``abjad.Pitch``.'
        interval = abs(self.note_heads[1].written_pitch
                       - self.note_heads[0].written_pitch).semitones
        sounding_pitch_dict = {1: 48,
                               2: 36,
                               3: 31,
                               4: 28,
                               5: 24,
                               7: 19,
                               9: 28,
                               12: 12,
                               16: 28,
                               19: 19,
                               24: 24,
                               28: 28,
                               }
        try:
            sounding_pitch = (self.note_heads[0].written_pitch
                              + sounding_pitch_dict[interval])
        except KeyError as err:
            raise ValueError('cannot calculate sounding pitch for given '
                             'interval') from err
        return sounding_pitch

    def sounding_note(self) -> abjad.Note:
        r'Returns the sounding note of the harmonic as an ``abjad.Note``.'
        note = abjad.Note(self.sounding_pitch(), self.written_duration)
        for indicator in abjad.inspect(self).indicators():
            abjad.attach(indicator, note)
        return note

    ### PUBLIC PROPERTIES ###

    @property
    def written_pitches(self) -> abjad.pitch.PitchSegment:
        r'The written pitches of the two note heads.'
        return abjad.pitch.PitchSegment(
            items=(note_head.written_pitch for note_head in self.note_heads),
            item_class=abjad.pitch.NamedPitch,
        )

    @written_pitches.setter
    def written_pitches(self,
                        written_pitches,
                        ):
        written_pitches_ = abjad.PitchSegment(written_pitches)
        if len(written_pitches_) != 2:
            raise ValueError("'ArtificialHarmonic' requires exactly two "
                             "pitches")
        for i, pitch in enumerate(written_pitches_):
            self.note_heads[i].written_pitch = pitch

    @property
    def style(self) -> str:
        r'The style of the upper note head.'
        return self._style

    @style.setter
    def style(self,
              style: str,
              ):
        if not isinstance(style, str):
            raise TypeError("'style' must be 'str'")
        self._style = style
        abjad.tweak(self.note_heads[1]).style = self._style

    @property
    def is_parenthesized(self) -> bool:
        r'Whether the bottom note head is parenthesised or not.'
        return self._is_parenthesized

    @is_parenthesized.setter
    def is_parenthesized(self,
                         is_parenthesized: bool,
                         ):
        if not isinstance(is_parenthesized, bool):
            raise TypeError("'is_parenthesized' must be 'bool'")
        self._is_parenthesized = is_parenthesized
        self.note_heads[0].is_parenthesized = self._is_parenthesized
        if self._is_parenthesized:
            abjad.tweak(self.note_heads[0]).ParenthesesItem__font_size = -4



from ._LoopParent import _LoopParent


class LoopByList(_LoopParent):
    r"""This class can be used to output slices of a ``list`` using the
    metaphor of a looping window of a variable size. This size is given by
    the argument ``window_size``, which is an ``int`` representing how many
    elements are to be included in each slice.

    For instance, if the initial container had the logical ties
    ``[A, B, C, D, E, F]`` (where each letter represents one elemnt of an
    arbitrary  type) and the looping window was size ``3``, the output would
    be:

    ``A B C B C D C D E D E F E F F``

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    ..  container:: example

        It takes a list and the number of elements of the window as arguments.
        Each call of the object, in this case ``looper()``, will move the
        window forwards and output the result:

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper()
        ['B', 'C', 'D']

        The property ``current_window`` can be used to access the current
        window without moving the head forwards.

        >>> looper.current_window
        ['B', 'C', 'D']

    ..  container:: example

        The very first call will output the input list without processing
        it. To disable this behaviour and have the looping window move on the
        very first call, initialise the class with the keyword argument
        ``move_window_on_first_call`` set to ``True``.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list,
        ...                            window_size=3,
        ...                            move_window_on_first_call=True,
        ...                            )
        >>> looper()
        ['B', 'C', 'D']

    ..  container:: example

        The instances of ``LoopByList`` can also be used as an iterator,
        which can then be used in a for loop to exhaust all windows.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list,
        ...                            window_size=3,
        ...                            )
        >>> for window in looper:
        ...     print(window)
        ['A', 'B', 'C']
        ['B', 'C', 'D']
        ['C', 'D', 'E']
        ['D', 'E', 'F']
        ['E', 'F']
        ['F']

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation. ``step_size`` dictates the size of each individual step in
        number of elements (default value is 1). ``max_steps`` sets the maximum
        number of steps that the window can advance when the object is called,
        ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition).
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window (default is 0).

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list,
        ...                            window_size=3,
        ...                            step_size=1,
        ...                            max_steps=2,
        ...                            repetition_chance=0.25,
        ...                            forward_bias=0.2,
        ...                            head_position=0,
        ...                            )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = 2
        >>> looper.step_size = 2
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 2
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2

    ..  container:: example

        The function ``len()`` can be used to get the total number of elements
        in the container.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> len(looper)
        6

    ..  container:: example

        To run through just part of the process and output it as a single
        list, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> looper.output_n(2)
        ['A', 'B', 'C', 'B', 'C', 'D']

    ..  container:: example

        To run through the whole process and output it as a single list, from
        the initial head position until the process outputs the single last
        element, use the method ``output_all()``.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> looper.output_all()
        ['A', 'B', 'C', 'B', 'C', 'D', 'C', 'D', 'D']

    .. container:: example

        To change the size of the looping window after instantiation, use the
        property ``window_size``. In the example below, the initial window is
        of size 3, and so the first call of the looper object outputs the
        first, second, and third elements of the list. The window size is then
        set to 4, and the looper is called again, moving to the element in the
        next position, thus outputting the second, third, fourth, and fifth
        elements.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper.window_size = 4
        >>> looper()
        ['B', 'C', 'D', 'E']

    .. container:: example

        Use the ``contents`` property to read as well as overwrite the contents
        of the looper. Notice that the ``head_position`` will remain on its
        previous value and must be reset to ``0`` if that's required.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopByList(input_list,
        ...                            window_size=3,
        ...                            )
        >>> looper.contents
        ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper()
        ['A', 'B', 'C']
        >>> looper()
        ['B', 'C', 'D']
        >>> looper.contents = [0, 1, 2, 3, 4]
        >>> looper.contents
        [0, 1, 2, 3, 4]
        >>> looper()
        [2, 3, 4]
        >>> looper.head_position = 0
        >>> looper()
        [0, 1, 2]

    .. container:: example

        It should be clear that the list can contain any types of elements:

        >>> input_list = [123, 'foo', (3, 4), 3.14]
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> looper()
        [123, 'foo', (3, 4)]

        This also include Abjad's types. Abjad's exclusive membership
        requirement is respected since each call returns a ``copy.deepcopy`` of
        the window. The same is true to the ``output_all()`` method.

        >>> import abjad
        >>> import copy
        >>> input_list = [
        ...     abjad.Container(r"c'4 d'4 e'4 f'4"),
        ...     abjad.Container(r"fs'1"),
        ...     abjad.Container(r"r2 bf4 c'4"),
        ...     abjad.Container(r"c''2. r4"),
        ... ]
        >>> looper = auxjad.LoopByList(input_list, window_size=3)
        >>> staff = abjad.Staff()
        >>> for element in looper.output_all():
        ...     staff.append(element)
        >>> abjad.f(staff)
        \new Staff
        {
            {
                c'4
                d'4
                e'4
                f'4
            }
            {
                fs'1
            }
            {
                r2
                bf4
                c'4
            }
            {
                fs'1
            }
            {
                r2
                bf4
                c'4
            }
            {
                c''2.
                r4
            }
            {
                r2
                bf4
                c'4
            }
            {
                c''2.
                r4
            }
            {
                c''2.
                r4
            }
        }

        .. figure:: ../_images/image-LoopByList-1.png
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: list,
                 *,
                 window_size: int,
                 step_size: int = 1,
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: int = 0,
                 move_window_on_first_call: bool = False,
                 ):
        self.contents = contents
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         move_window_on_first_call,
                         )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(self._contents)

    def __len__(self) -> int:
        r'Outputs a length of ``contents``.'
        return len(self._contents)

    def __call__(self) -> list:
        r"""Calls the looping process for one iteration, returning a ``list``.

        This method replaces the parent's one since the parent's method outputs
        an ``abjad.Selection``.
        """
        self._move_head()
        if self._done:
            raise RuntimeError("'contents' has been exhausted")
        self._slice_contents()
        return self.current_window[:]

    def __next__(self) -> list:
        r"""Calls the looping process for one iteration, returning a ``list``.

        This method replaces the parent's one since the parent's method outputs
        an ``abjad.Selection``.
        """
        self._move_head()
        if self._done:
            raise StopIteration
        self._slice_contents()
        return self._current_window[:]

    ### PUBLIC METHODS ###

    def output_all(self) -> list:
        r"""Goes through the whole looping process and outputs a single list.
        This method replaces the parent's one since the parent's method outputs
        an ``abjad.Selection``.
        """
        dummy_container = []
        while True:
            try:
                dummy_container.extend(self.__call__())
            except RuntimeError:
                break
        return dummy_container[:]

    def output_n(self, n: int) -> list:
        r"""Goes through ``n`` iterations of the looping process and outputs a
        single list. This method replaces the parent's one since the parent's
        method outputs an ``abjad.Selection``.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 0:
            raise ValueError("argument must be a positive 'int'")
        dummy_container = []
        for _ in range(n):
            dummy_container.extend(self.__call__())
        return dummy_container[:]

    ### PRIVATE METHODS ###

    def _slice_contents(self):
        r"""This method takes a slice with ``window_size`` number of elements
        out of the contents starting at the current ``head_position``.
        """
        start = self.head_position
        end = self.head_position + self.window_size
        self._current_window = self._contents[start:end]

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self):
        r'The ``list`` which serves as the basis for the slices of the looper.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: list,
                 ):
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list")
        self._contents = contents[:]



import abjad
import pytest
import auxjad


def test_LoopByWindow_01():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        """)
    notes = looper.current_window
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        """)


def test_LoopByWindow_02():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'2
            e'4
        }
        """)


def test_LoopByWindow_03():
    input_music = abjad.Container(r"c'4 d'2 e'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 8),
                                 )
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'2
            e'4
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4.
            e'8
            ~
            e'8
            r8
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4
            e'4
            r4
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'8
            e'8
            ~
            e'8
            r4.
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            e'4
            r2
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            e'8
            r8
            r2
        }
        """)
    with pytest.raises(StopIteration):
        assert looper.__next__()


def test_LoopByWindow_04():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(5, 8),
                                 max_steps=2,
                                 repetition_chance=0.25,
                                 forward_bias=0.2,
                                 head_position=(2, 8),
                                 omit_time_signature=False,
                                 )
    assert looper.window_size == abjad.Meter((3, 4))
    assert looper.step_size == abjad.Duration((5, 8))
    assert looper.max_steps == 2
    assert looper.repetition_chance == 0.25
    assert looper.forward_bias == 0.2
    assert looper.head_position == abjad.Duration((1, 4))
    assert not looper.omit_time_signature
    looper.window_size = (5, 4)
    looper.step_size = (1, 4)
    looper.max_steps = 3
    looper.repetition_chance = 0.1
    looper.forward_bias = 0.8
    looper.head_position = 0
    looper.omit_time_signature = True
    assert looper.window_size == abjad.Meter((5, 4))
    assert looper.step_size == abjad.Duration((1, 4))
    assert looper.max_steps == 3
    assert looper.repetition_chance == 0.1
    assert looper.forward_bias == 0.8
    assert looper.head_position == abjad.Duration(0)
    assert looper.omit_time_signature


def test_LoopByWindow_05():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    music = looper.output_all()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
            e'4
            f'4
            r4
            f'4
            r2
        }
        """)


def test_LoopByWindow_06():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
            ~
            e'8
            f'8
        }
        """)
    looper.window_size = (3, 8)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/8
            c'16
            d'16
            ~
            d'4
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4.
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4.
        }
        """)


def test_LoopByWindow_07():
    input_music = abjad.Container(r"\times 2/3 {c'8 d'8 e'} d'2.")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 16))
    staff = abjad.Staff()
    for _ in range(3):
        window = looper()
        staff.append(window)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \times 2/3 {
                \time 3/4
                c'8
                d'8
                e'8
            }
            d'2
            \times 2/3 {
                c'32
                d'16
                ~
                d'16
                e'8
            }
            d'16
            ~
            d'2
            \times 2/3 {
                d'16
                e'8
            }
            d'8
            ~
            d'2
        }
        """)


def test_LoopByWindow_08():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music, omit_time_signature=True)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'4
            d'2
            e'4
        }
        """)


def test_LoopByWindow_09():
    wrong_type_input = 'foobar'
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    with pytest.raises(TypeError):
        assert auxjad.LoopByWindow(wrong_type_input)
        assert auxjad.LoopByWindow(input_music, window_size=17j)
        assert auxjad.LoopByWindow(input_music, max_steps='foobar')
        assert auxjad.LoopByWindow(input_music, repetition_chance='foobar')
        assert auxjad.LoopByWindow(input_music, head_position=62.3j)
        assert auxjad.LoopByWindow(input_music, omit_time_signature='foobar')
    with pytest.raises(ValueError):
        assert auxjad.LoopByWindow(input_music, window_size=(100, 1))
        assert auxjad.LoopByWindow(input_music, max_steps=-1)
        assert auxjad.LoopByWindow(input_music, repetition_chance=-0.3)
        assert auxjad.LoopByWindow(input_music, repetition_chance=1.4)
        assert auxjad.LoopByWindow(input_music, head_position=(100, 1))


def test_LoopByWindow_10():
    input_music = abjad.Container(r"c'4 e'2 d'2 f'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            e'2
            ~
            e'2
            d'4
            e'4
            d'2
            ~
            d'2
            f'4
            d'4
            f'4
            r4
            f'4
            r2
        }
        """)


def test_LoopByWindow_11():
    input_music = abjad.Container(r"c'4 <e' f' g'>2 r4 f'2.")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            <e' f' g'>2
            ~
            <e' f' g'>2
            r4
            <e' f' g'>4
            r4
            f'4
            r4
            f'2
            ~
            f'2.
            ~
            f'2
            r4
            f'4
            r2
        }
        """)


def test_LoopByWindow_12():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    music = looper.output_n(2)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
        }
        """)


def test_LoopByWindow_13():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    music = looper.output_n(2, tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'2
            ~
            d'2
            e'4
        }
        """)


def test_LoopByWindow_14():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 )
    with pytest.raises(RuntimeError):
        looper.output_n(100)


def test_LoopByWindow_15():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 head_position=(3, 4),
                                 forward_bias=0.0,
                                 )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            f'4
            g'4
            a'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            e'4
            f'4
            g'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4
            e'4
            f'4
        }
        """)


def test_LoopByWindow_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 head_position=(3, 4),
                                 forward_bias=0.0,
                                 )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 3/4
        f'4
        g'4
        a'4
        e'4
        f'4
        g'4
        d'4
        e'4
        f'4
        c'4
        d'4
        e'4
    }
    """)


def test_LoopByWindow_17():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopByWindow(input_music,
                                 window_size=(3, 4),
                                 step_size=(1, 4),
                                 forward_bias=0.0,
                                 )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 3/4
        c'4
        d'4
        e'4
    }
    """)


def test_LoopByWindow_18():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music,
                                 move_window_on_first_call=True,
                                 )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 4/4
        c'8.
        d'16
        ~
        d'4..
        e'16
        ~
        e'8.
        f'16
    }
    """)


def test_LoopByWindow_19():
    input_music = abjad.Container(r"c'4-.\pp\< d'2--\f e'4->\ppp f'2 ~ f'8")
    looper = auxjad.LoopByWindow(input_music)
    staff = abjad.Staff()
    for _ in range(2):
        music = looper()
        staff.append(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 4/4
        c'4
        \pp
        - \staccato
        \<
        d'2
        \f
        - \tenuto
        e'4
        \ppp
        - \accent
        c'8.
        \pp
        - \staccato
        \<
        d'16
        \f
        - \tenuto
        ~
        d'4..
        e'16
        \ppp
        - \accent
        ~
        e'8.
        f'16
    }
    """)


def test_LoopByWindow_20():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByWindow(input_music)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        """)
    looper.contents = abjad.Container(r"c'16 d'16 e'16 f'16 g'2. | a'1")
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            e'16
            f'16
            g'8
            ~
            g'2
            ~
            g'8
            a'8
        }
        """)
    looper.head_position = 0
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'16
            d'16
            e'16
            f'16
            g'2.
        }
        """)


def test_LoopByWindow_21():
    input_music = abjad.Container(r"c'1")
    looper = auxjad.LoopByWindow(input_music)
    assert len(looper) == 16
    input_music = abjad.Container(r"c'1")
    looper = auxjad.LoopByWindow(input_music,
                                 step_size=(1, 4),
                                 )
    assert len(looper) == 4
    input_music = abjad.Container(r"c'2..")
    looper = auxjad.LoopByWindow(input_music,
                                 step_size=(1, 4),
                                 window_size=(2, 4),
                                 )
    assert len(looper) == 4



"""auxjad
======

``auxjad`` is a library of auxiliary functions and classes for Abjad 3.1. All
classes and functions have a __doc__ attribute with usage instructions.

Documentation is available at https://gilbertohasnofb.github.io/auxjad-docs/. A
pdf version of the documentation is also available in the ``/docs`` directory
of the repository.

Bugs can be reported through the project's Issue Tracker.

This library is published under the MIT License.
"""

from .core import *
from .entities import *
from .utilities import *


__author__ = "Gilberto Agostinho <gilbertohasnofb@gmail.com>"
__version__ = "0.7.0"



import copy
import abjad
from .close_container import close_container
from .container_is_full import container_is_full
from .remove_repeated_time_signatures import remove_repeated_time_signatures


def repeat_container(container: abjad.Container,
                     n: int,
                     *,
                     omit_all_time_signatures: bool = False,
                     force_identical_time_signatures: bool = False,
                     ) -> abjad.Container:
    r"""This Returns an ``abjad.Container`` with n repetitions of an input
    container (of type ``abjad.Container`` or child class).

    ..  container:: example

        The required arguments are an ``abjad.Container`` (or child class) and
        and integer n for the number of repetitions.

        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> output_container = auxjad.repeat_container(container, 3)
        >>> abjad.f(output_container)
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }

        .. figure:: ../_images/image-repeat_container-1.png

    .. note::

        Notice that the time signatures in the output are commented out with
        ``%%%``. This is because Abjad only applies time signatures to
        containers that belong to a ``abjad.Staff``. The present function works
        with either ``abjad.Container`` and ``abjad.Staff``.
        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> output_container = auxjad.repeat_container(container, 3)
        >>> abjad.f(output_container)
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }
        >>> staff = abjad.Staff([output_container])
        >>> abjad.f(output_container)
        {
            \time 3/4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }

    ..  container:: example

        It handle containers with multiple bars and different time signatures.

        >>> container = abjad.Container(r"\time 3/4 c'2. \time 2/4 r2 g'2")
        >>> output_container = auxjad.repeat_container(container, 3)
        >>> abjad.f(output_container)
        {
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
        }

        .. figure:: ../_images/image-repeat_container-4.png

    ..  container:: example

        It automatically closes a container if necessary.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'2")
        >>> output_container = auxjad.repeat_container(container, 2)
        >>> abjad.f(output_container)
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            %%% \time 2/4 %%%
            f'2
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            %%% \time 2/4 %%%
            f'2
        }

        .. figure:: ../_images/image-repeat_container-5.png

    ..  container:: example

        To omit all time signatures, set the keyword argument
        ``omit_all_time_signatures`` to ``True``.

        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> output_container = auxjad.repeat_container(
        ...     container,
        ...     3,
        ...     omit_all_time_signatures=True,
        ... )
        >>> abjad.f(output_container)
        {
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }

        .. figure:: ../_images/image-repeat_container-6.png

    ..  container:: example

        To force identical time signatures to be repeated at every repetition,
        set the keyword argument  ``force_identical_time_signatures`` to
        ``True``.

        >>> container = abjad.Container(r"\time 5/4 c'2. d'4 e'4")
        >>> output_container = auxjad.repeat_container(
        ...     container,
        ...     3,
        ...     force_identical_time_signatures=True,
        ... )
        >>> abjad.f(output_container)
        {
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
        }

        .. figure:: ../_images/image-repeat_container-7.png

    ..  container:: example

        The input container can be of child classes such as ``abjad.Staff``,
        and the output will be of the same type.

        >>> container = abjad.Staff(r"c'4 d'4 e'4")
        >>> output_staff = auxjad.repeat_container(container, 3)
        >>> abjad.f(output_staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }

        .. figure:: ../_images/image-repeat_container-8.png

    ..  warning::

        If a container is malformed, i.e. it has an underfilled bar before a
        time signature change, the function raises a ``ValueError`` exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'4")
        >>> output_container = auxjad.repeat_container(container)
        ValueError: 'container' is malformed, with an underfull bar preceeding
        a time signature change
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first positional argument must be 'abjad.Container' "
                        "or child class")
    if not isinstance(n, int):
        raise TypeError("second positional argument must be 'int'")
    if not isinstance(omit_all_time_signatures, bool):
        raise TypeError("'omit_all_time_signatures' must be 'bool'")
    if not isinstance(force_identical_time_signatures, bool):
        raise TypeError("'force_identical_time_signatures' must be 'bool'")

    container_ = copy.deepcopy(container)
    if not container_is_full(container_):
        close_container(container_)
    output_container = type(container_)()
    for _ in range(n):
        output_container.extend(copy.deepcopy(container_))
    if not force_identical_time_signatures:
        remove_repeated_time_signatures(output_container)
    if omit_all_time_signatures:
        for leaf in abjad.select(output_container).leaves():
            if abjad.inspect(leaf).indicator(abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)
    return output_container



import abjad
import pytest
import auxjad


def test_HarmonicNote_01():
    harm = auxjad.HarmonicNote("c''4")
    assert harm.style == 'harmonic'
    assert format(harm) == abjad.String.normalize(
        r"""
        \tweak style #'harmonic
        c''4
        """)


def test_HarmonicNote_02():
    harm1 = auxjad.HarmonicNote("c''4")
    harm2 = auxjad.HarmonicNote("c''", 1/4)
    harm3 = auxjad.HarmonicNote(12, 0.25)
    harm4 = auxjad.HarmonicNote(12, abjad.Duration(1, 4))
    harms = [harm1, harm2, harm3, harm4]
    for harm in harms:
        assert format(harm) == abjad.String.normalize(
            r"""
            \tweak style #'harmonic
            c''4
            """)


def test_HarmonicNote_03():
    harm = auxjad.HarmonicNote("c''4",
                               style='harmonic-mixed',
                               )
    assert harm.style == 'harmonic-mixed'
    assert format(harm) == abjad.String.normalize(
        r"""
        \tweak style #'harmonic-mixed
        c''4
        """)


def test_HarmonicNote_04():
    harm = auxjad.HarmonicNote("c''4",
                               multiplier=(2, 3),
                               )
    assert harm.multiplier == abjad.Multiplier(2, 3)
    assert format(harm) == abjad.String.normalize(
        r"""
        \tweak style #'harmonic
        c''4 * 2/3
        """)


def test_HarmonicNote_05():
    harm = auxjad.HarmonicNote("c''4")
    assert harm.written_pitch == "c''"
    assert harm.written_duration == 1/4
    assert harm.style == 'harmonic'
    harm.written_pitch = 18
    harm.written_duration = abjad.Duration(1, 8)
    harm.style = 'harmonic-mixed'
    assert harm.written_pitch == "fs''"
    assert harm.written_duration == 1/8
    assert harm.style == 'harmonic-mixed'


def test_HarmonicNote_06():
    harm = auxjad.HarmonicNote("c''1",
                               style='flageolet',
                               )
    assert harm.style == 'flageolet'
    assert format(harm) == abjad.String.normalize(
        r"""
        c''1
        \flageolet
        """)


def test_HarmonicNote_07():
    harm1 = auxjad.HarmonicNote("d''1")
    harm2 = auxjad.HarmonicNote("d''1",
                                markup='III.',
                                )
    harm3 = auxjad.HarmonicNote("d''1",
                                markup='III.',
                                direction=abjad.Down)
    staff = abjad.Staff([harm1, harm2, harm3])
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \tweak style #'harmonic
            d''1
            \tweak style #'harmonic
            d''1
            ^ \markup { III. }
            \tweak style #'harmonic
            d''1
            _ \markup { III. }
        }
        """)


def test_HarmonicNote_08():
    harm = auxjad.HarmonicNote("d''1",
                               markup='III.',
                               )
    harm.markup = None
    assert format(harm) == abjad.String.normalize(
        r"""
        \tweak style #'harmonic
        d''1
        """)


def test_HarmonicNote_09():
    harm = auxjad.HarmonicNote("d''1")
    abjad.attach(abjad.Markup('test'), harm)
    harm.markup = 'III.'
    with pytest.raises(Exception):
        harm.markup = None



import abjad
import pytest
import auxjad


def test_LoopByNotes_01():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music, window_size=3)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }
        """)
    notes = looper.current_window
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }
        """)


def test_LoopByNotes_02():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=3,
                                step_size=1,
                                max_steps=2,
                                repetition_chance=0.25,
                                forward_bias=0.2,
                                head_position=0,
                                omit_all_time_signatures=False,
                                force_identical_time_signatures=False,
                                )
    assert looper.window_size == 3
    assert looper.step_size == 1
    assert looper.max_steps == 2
    assert looper.repetition_chance == 0.25
    assert looper.forward_bias == 0.2
    assert looper.head_position == 0
    assert not looper.omit_all_time_signatures
    assert not looper.force_identical_time_signatures
    looper.window_size = 2
    looper.step_size = 2
    looper.max_steps = 3
    looper.repetition_chance = 0.1
    looper.forward_bias = 0.8
    looper.head_position = 2
    looper.omit_all_time_signatures = True
    looper.force_identical_time_signatures = True
    assert looper.window_size == 2
    assert looper.step_size == 2
    assert looper.max_steps == 3
    assert looper.repetition_chance == 0.1
    assert looper.forward_bias == 0.8
    assert looper.head_position == 2
    assert looper.omit_all_time_signatures
    assert looper.force_identical_time_signatures


def test_LoopByNotes_03():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music, window_size=3)
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 1
    looper()
    assert looper.head_position == 2


def test_LoopByNotes_04():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music, window_size=3)
    assert len(looper) == 5


def test_LoopByNotes_05():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            d'4
            e'4
            e'4
            f'4
            \time 1/4
            f'4
        }
        """)


def test_LoopByNotes_06():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music, window_size=3)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        """)
    looper.window_size = 4
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 19/8
            d'2
            e'4
            f'2
            ~
            f'8
            g'1
        }
        """)


def test_LoopByNotes_07():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=3,
                                head_position=2,
                                )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 15/8
            e'4
            f'2
            ~
            f'8
            g'1
        }
        """)


def test_LoopByNotes_08():
    input_music = abjad.Container(r"c'4 d'8 \times 2/3 {a4 g2}")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            #(ly:expect-warning "strange time signature found")
            \time 7/24
            d'8
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                \time 2/4
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                g2
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                #(ly:expect-warning "strange time signature found")
                \time 2/6
                g2
            }
        }
        """)


def test_LoopByNotes_09():
    input_music = abjad.Container(r"c'4 d'2 e'4")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'2
            e'4
        }
        """)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 1/4
            e'4
        }
        """)
    with pytest.raises(StopIteration):
        assert looper.__next__()


def test_LoopByNotes_10():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=3,
                                omit_all_time_signatures=True,
                                )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'4
            d'2
            e'4
        }
        """)


def test_LoopByNotes_11():
    wrong_type_input = 'foo'
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    with pytest.raises(TypeError):
        assert auxjad.LoopByNotes(wrong_type_input, window_size=3)
        assert auxjad.LoopByNotes(input_music, window_size='foobar')
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  step_size='foobar',
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  max_steps='foobar',
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  repetition_chance='foobar',
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  head_position='foobar',
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  omit_all_time_signatures='foobar',
                                  )
    with pytest.raises(ValueError):
        assert auxjad.LoopByNotes(input_music, window_size=100)
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  step_size=-1,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  step_size=100,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  max_steps=-1,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  repetition_chance=-0.3,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  repetition_chance=1.4,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  head_position=-1,
                                  )
        assert auxjad.LoopByNotes(input_music,
                                  window_size=3,
                                  head_position=100,
                                  )


def test_LoopByNotes_12():
    input_music = abjad.Container(r"c'4 d'2 r8 d'4 <e' g'>8 r4 f'2. <e' g'>16")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=4,
                                )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 9/8
            c'4
            d'2
            r8
            d'4
            ~
            \time 4/4
            d'2
            r8
            d'4
            <e' g'>8
            \time 3/4
            r8
            d'4
            <e' g'>8
            r4
            \time 11/8
            d'4
            <e' g'>8
            r4
            f'2.
            \time 19/16
            <e' g'>8
            r4
            f'2.
            <e' g'>16
            \time 17/16
            r4
            f'2.
            <e' g'>16
            \time 13/16
            f'2.
            <e' g'>16
            ~
            \time 1/16
            <e' g'>16
        }
        """)


def test_LoopByNotes_13():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    notes = looper.output_n(2)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            d'4
            e'4
        }
        """)


def test_LoopByNotes_14():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    notes = looper.output_n(2, tie_identical_pitches=True)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            ~
            d'4
            e'4
        }
        """)


def test_LoopByNotes_15():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music, window_size=2)
    with pytest.raises(RuntimeError):
        looper.output_n(100)


def test_LoopByNotes_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=2,
                                head_position=2,
                                forward_bias=0.0,
                                )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            e'4
            f'4
            d'4
            e'4
            c'4
            d'4
        }
        """)


def test_LoopByNotes_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=2,
                                head_position=0,
                                forward_bias=0.0,
                                )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 2/4
        c'4
        d'4
    }
    """)


def test_LoopByNotes_17():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=2,
                                force_identical_time_signatures=True,
                                )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
            \time 2/4
            e'4
            f'4
            \time 1/4
            f'4
        }
        """)


def test_LoopByNotes_18():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=3,
                                move_window_on_first_call=True,
                                )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }
        """)


def test_LoopByNotes_19():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopByNotes(input_music,
                                window_size=3,
                                )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
        }
        """)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4
            e'4
            f'4
        }
        """)
    looper.contents = abjad.Container(r"c'''4 r4 d'''4 r4 e'''4 r4 f'''4 r4")
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'''4
            r4
            e'''4
        }
        """)
    looper.head_position = 0
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'''4
            r4
            d'''4
        }
        """)



import copy
import random
import abjad
from ..utilities.remove_repeated_time_signatures import (
    remove_repeated_time_signatures
)
from ..utilities.simplified_time_signature_ratio import (
    simplified_time_signature_ratio
)


class Shuffler:
    r"""Shuffler takes an input ``abjad.Container`` and shuffles its logical
    ties. It can also shuffle only pitches, as well as rotate them. When
    shuffling or rotating pitches only, tuplets are allowed. Tuplets are not
    supported when shuffling leaves.

    ..  container:: example

        Calling the object will output a shuffled selection of the input
        container.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            d'4
            c'4
            f'4
            e'4
        }

        .. figure:: ../_images/image-Shuffler-1.png

        To get the result of the last operation, use the property
        ``current_window``.

        >>> music = shuffler.current_window
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            d'4
            c'4
            f'4
            e'4
        }

        .. figure:: ../_images/image-Shuffler-2.png

    ..  container:: example

        Calling the object outputs the same result as using the method
        ``shuffle()``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.shuffle()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            f'4
            c'4
            e'4
            d'4
        }

        .. figure:: ../_images/image-Shuffler-3.png

    ..  container:: example

        This class has many keyword arguments, all of which can be altered
        after instantiation using properties with the same names as shown
        below.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 |"
        ...                             r"\time 2/4 f'4 g'4 |"
        ...                             )
        >>> shuffler = auxjad.Shuffler(container,
        ...                            output_single_measure=False,
        ...                            disable_rewrite_meter=False,
        ...                            force_time_signatures=False,
        ...                            omit_time_signatures=False,
        ...                            )
        >>> shuffler.output_single_measure
        False
        >>> shuffler.disable_rewrite_meter
        False
        >>> shuffler.force_time_signatures
        False
        >>> shuffler.omit_time_signatures
        False
        >>> shuffler.output_single_measure = True
        >>> shuffler.disable_rewrite_meter = True
        >>> shuffler.force_time_signatures = True
        >>> shuffler.omit_time_signatures = True
        >>> shuffler.output_single_measure
        True
        >>> shuffler.disable_rewrite_meter
        True
        >>> shuffler.force_time_signatures
        True
        >>> shuffler.omit_time_signatures
        True


    ..  container:: example

        If ``output_single_measure`` is set to ``True``, then the whole
        container is output as a single measure, having its time signature
        rewritten.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 |"
        ...                             r"\time 2/4 f'4 g'4"
        ...                             )
        >>> shuffler = auxjad.Shuffler(container,
        ...                            output_single_measure=True,
        ...                            )
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 5/4
            f'4
            d'4
            e'4
            g'4
            c'4
        }

        .. figure:: ../_images/image-Shuffler-4.png

    ..  container:: example

        If ``disable_rewrite_meter`` is set to ``True``, then the automatic
        behaviour of rewriting the leaves according to the meter is disabled.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 |"
        ...                             r"\time 2/4 f'2"
        ...                             )
        >>> shuffler = auxjad.Shuffler(container,
        ...                            output_single_measure=True,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 5/4
            d'4..
            f'2
            c'16
            e'4
        }

        .. figure:: ../_images/image-Shuffler-5.png

    ..  container:: example

        The first call to the instance will add the correct time signature to
        the first leaf. Subsequent calls will only add it if its necessary,
        such as when there is a time signature change in some bar in the
        container.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            d'4..
            e'16
            ~
            e'8.
            f'16
            ~
            f'4..
            r16
            r8.
            c'16
        }

        .. figure:: ../_images/image-Shuffler-6.png

        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            c'16
            e'8.
            ~
            e'16
            f'4..
            ~
            f'16
            r8.
            r16
            d'4..
        }

        .. figure:: ../_images/image-Shuffler-7.png

        It is possible to force time signatures on every call using either
        optional keyword argument ``force_time_signatures``.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            force_time_signatures=True,
        ...                            )
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            d'4..
            r16
            r8.
            c'16
            f'2
            e'4
        }

        .. figure:: ../_images/image-Shuffler-8.png

        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'16
            e'8.
            ~
            e'16
            f'4..
            ~
            f'16
            r8.
            r16
            d'4..
        }

        .. figure:: ../_images/image-Shuffler-9.png

    ..  container:: example

        To disable time signatures altogether, initialise this class with the
        keyword argument ``omit_time_signatures`` set to ``True`` (default is
        ``False``), or change the ``omit_time_signatures`` property after
        initialisation.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            omit_time_signatures=True,
        ...                            )
        >>> music = shuffler()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            d'4..
            e'16
            ~
            e'8.
            f'16
            ~
            f'4..
            r16
            r8.
            c'16
        }

        .. figure:: ../_images/image-Shuffler-10.png

        >>> shuffler.omit_time_signatures
        True
        >>> shuffler.omit_time_signatures = False
        >>> shuffler.omit_time_signatures
        False

    ..  container:: example

        To output several shuffled containers at once, use the ``output_n``
        method, inputting the desired number of iterations.

        >>> container = abjad.Container(r"\time 2/4 c'16 d'4.. | r4 e'8. f'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.output_n(3)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/4
            d'4..
            f'16
            c'16
            e'8.
            r4
            d'4..
            e'16
            ~
            e'8
            f'16
            r16
            r8.
            c'16
            r4
            d'4
            ~
            d'8.
            f'16
            c'16
            e'8.
        }

        .. figure:: ../_images/image-Shuffler-11.png

    ..  container:: example

        To shuffle only pitches, keeping the durations of the leaves as they
        are, use the method ``shuffle_pitches()``. It handles both notes and
        chords. Rests will remain at their current location.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.shuffle_pitches()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            e'16
            c'4..
            r4
            d'8.
            f'16
        }

        .. figure:: ../_images/image-Shuffler-12.png

    ..  container:: example

        When dealing with pitches, it is possible to use containers containing
        tuplets. And similarly to the method ``output_n()``, to output several
        containers with shuffled pitches, use ``output_n_shuffled_pitches()``.

        >>> container = abjad.Container(r"\times 2/3 {\time 5/4 c'4 d'2}"
        ...                             r"r4 e'4. f'8"
        ...                             )
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.output_n_shuffled_pitches(3)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \times 2/3 {
                \time 5/4
                f'4
                e'2
            }
            r4
            d'4.
            c'8
            \times 2/3 {
                d'4
                c'2
            }
            r4
            f'4.
            e'8
            \times 2/3 {
                d'4
                f'2
            }
            r4
            c'4.
            e'8
        }

        .. figure:: ../_images/image-Shuffler-13.png

    ..  container:: example

        To rotate pitches, use the ``rotate_pitches()`` method.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.rotate_pitches()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            d'16
            e'4..
            r4
            f'8.
            c'16
        }

        .. figure:: ../_images/image-Shuffler-14.png

    ..  container:: example

        This method can take two optioanl keyword arguments: ``anticlockwise``,
        set to ``False`` by default, and ``n_rotations``, set to 1 by default.
        The first defines the direction of the rotation, while the later sets
        the number of rotations applied.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.rotate_pitches(anticlockwise=True, n_rotations=2)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            e'16
            f'4..
            r4
            c'8.
            d'16
        }

        .. figure:: ../_images/image-Shuffler-15.png

    ..  container:: example

        Similarly to the method ``output_n()``, to output several containers
        with rotated pitches, use ``output_n_rotated_pitches()``.

        >>> container = abjad.Container(r"\times 2/3 {\time 5/4 c'4 d'2}"
        ...                             r"r4 e'4. f'8"
        ...                             )
        >>> shuffler = auxjad.Shuffler(container)
        >>> music = shuffler.output_n_rotated_pitches(3)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \times 2/3 {
                \time 5/4
                d'4
                e'2
            }
            r4
            f'4.
            c'8
            \times 2/3 {
                e'4
                f'2
            }
            r4
            c'4.
            d'8
            \times 2/3 {
                f'4
                c'2
            }
            r4
            d'4.
            e'8
        }

        .. figure:: ../_images/image-Shuffler-16.png

    ..  container:: example

        Use the property ``contents`` to get the input container upon which the
        shuffler operates. Notice that ``contents`` remains invariant after
        any shuffling or rotation operations (use ``current_window`` for the
        transformed selection of music). ``contents`` can be used to change the
        ``abjad.Container`` to be shuffled.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> shuffler = auxjad.Shuffler(container)
        >>> abjad.f(shuffler.contents)
        {
            c'4
            d'4
            e'4
            f'4
        }

        .. figure:: ../_images/image-Shuffler-17.png

        >>> shuffler()
        >>> abjad.f(shuffler.contents)
        {
            c'4
            d'4
            e'4
            f'4
        }

        .. figure:: ../_images/image-Shuffler-18.png

        >>> shuffler.contents = abjad.Container(r"cs2 ds2")
        >>> abjad.f(shuffler.contents)
        {
            cs2
            ds2
        }

        .. figure:: ../_images/image-Shuffler-19.png
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: abjad.Container,
                 *,
                 output_single_measure: bool = False,
                 disable_rewrite_meter: bool = False,
                 force_time_signatures: bool = False,
                 omit_time_signatures: bool = False,
                 ):
        self.contents = contents
        self._update_logical_ties()
        self._find_time_signatures()
        self.output_single_measure = output_single_measure
        self.disable_rewrite_meter = disable_rewrite_meter
        self.force_time_signatures = force_time_signatures
        self.omit_time_signatures = omit_time_signatures
        self._last_time_signature = None

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(abjad.f(self._current_window))

    def __len__(self) -> int:
        r'Outputs the number of logical ties of ``contents``.'
        return len(self._logical_ties)

    def __call__(self) -> abjad.Selection:
        r'Calls the shuffling process, returning an ``abjad.Container``'
        return self.shuffle()

    ### PUBLIC METHODS ###

    def shuffle(self) -> abjad.Selection:
        r'Shuffles the logical ties of ``contents``.'
        if self._force_time_signatures:
            self._last_time_signature = None
        dummy_container = abjad.Container()
        indeces = list(range(self.__len__()))
        random.shuffle(indeces)
        for index in indeces:
            logical_tie = copy.deepcopy(
                self._logical_ties[index])
            dummy_container.append(logical_tie)
            for leaf in logical_tie:
                if abjad.inspect(leaf).effective(abjad.TimeSignature):
                    abjad.detach(abjad.TimeSignature, leaf)
        if not self._output_single_measure:
            # splitting leaves at bar line points
            abjad.mutate(dummy_container[:]).split(
                self._time_signatures_durations,
                cyclic=True,
            )
            # attaching time signature structure
            time_signature = self._time_signatures[0]
            if (not self._last_time_signature
                    or time_signature != self._last_time_signature):
                abjad.attach(time_signature, dummy_container[0])
                self._last_time_signature = time_signature
            duration = abjad.inspect(dummy_container[0]).duration()
            index = 1
            for leaf in abjad.select(dummy_container).leaves()[1:]:
                if duration % self._time_signatures[index - 1].duration == 0:
                    time_signature = self._time_signatures[index]
                    if time_signature != self._last_time_signature:
                        abjad.attach(time_signature, leaf)
                        self._last_time_signature = time_signature
                    if index + 1 < len(self._time_signatures):
                        index += 1
                        duration = abjad.Duration(0)
                    else:
                        break
                duration += abjad.inspect(leaf).duration()
            remove_repeated_time_signatures(dummy_container)
            self._last_time_signature = self._time_signatures[-1]
        else:
            time_signature = abjad.TimeSignature(
                abjad.inspect(dummy_container).duration())
            time_signature = simplified_time_signature_ratio(time_signature)
            if (not self._last_time_signature
                    or time_signature != self._last_time_signature):
                abjad.attach(time_signature, dummy_container[0])
            self._last_time_signature = time_signature
        # rewrite meter
        if not self._disable_rewrite_meter:
            start = 0
            duration = abjad.Duration(0)
            index = 0
            dummy_container_leaves = abjad.select(dummy_container).leaves()
            for leaf_n in range(len(dummy_container_leaves)):
                duration = abjad.inspect(
                    dummy_container_leaves[start : leaf_n+1]).duration()
                if duration == self._time_signatures_durations[index]:
                    abjad.mutate(
                        dummy_container_leaves[start : leaf_n+1]
                    ).rewrite_meter(self._time_signatures[index])
                    if index + 1 < len(self._time_signatures):
                        index += 1
                        start = leaf_n + 1
                    else:
                        break
        # output
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        return self.current_window

    def shuffle_pitches(self) -> abjad.Selection:
        r'Shuffles only the pitches of ``contents``.'
        if self._force_time_signatures:
            self._last_time_signature = None
        pitches = self._get_pitch_list()
        # shuffling (while preserving rests)
        true_pitches = [pitch for pitch in pitches if pitch is not None]
        random.shuffle(true_pitches)
        index = 0
        for true_pitch in true_pitches:
            while not pitches[index]:
                index += 1
            pitches[index] = true_pitch
            index += 1
        # rewriting leaves
        self._rewrite_pitches(pitches)
        self._last_time_signature = self._time_signatures[-1]
        self._update_logical_ties()
        # updating logical ties
        return self.current_window

    def rotate_pitches(self,
                       *,
                       n_rotations: int = 1,
                       anticlockwise: bool = False,
                       ) -> abjad.Selection:
        r'Rotates the pitches of ``contents``.'
        if not isinstance(n_rotations, int):
            raise TypeError("'n_rotations' must be 'int'")
        if n_rotations < 1:
            raise ValueError("'n_rotations' must be greater than zero")
        if not isinstance(anticlockwise, bool):
            raise TypeError("'anticlockwise' must be 'bool'")
        if self._force_time_signatures:
            self._last_time_signature = None
        pitches = self._get_pitch_list()
        # rotating pitches (while preserving rests)
        true_pitches = [pitch for pitch in pitches if pitch is not None]
        for _ in range(n_rotations):
            if not anticlockwise:
                true_pitches = true_pitches[1:] + [true_pitches[0]]
            else:
                true_pitches = [true_pitches[-1]] + true_pitches[:-1]
        index = 0
        for true_pitch in true_pitches:
            while not pitches[index]:
                index += 1
            pitches[index] = true_pitch
            index += 1
        # rewriting leaves
        self._rewrite_pitches(pitches)
        self._last_time_signature = self._time_signatures[-1]
        # updating logical ties
        return self.current_window

    def output_n(self,
                 n: int,
                 ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the shuffling process and outputs
        a single ``abjad.Selection``.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 1:
            raise ValueError("argument must be greater than zero")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.__call__())
        remove_repeated_time_signatures(dummy_container)
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    def output_n_shuffled_pitches(self,
                                  n: int,
                                  ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the pitch shuffling process and
        outputs a single ``abjad.Selection``.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 1:
            raise ValueError("argument must be greater than zero")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.shuffle_pitches())
        remove_repeated_time_signatures(dummy_container)
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    def output_n_rotated_pitches(self,
                                 n: int,
                                 *,
                                 n_rotations: int = 1,
                                 anticlockwise: bool = False,
                                 ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the pitch rotation process and
        outputs a single ``abjad.Selection``.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive "
                             "'int'")
        if not isinstance(n_rotations, int):
            raise TypeError("'n_rotations' must be 'int'")
        if not isinstance(anticlockwise, bool):
            raise TypeError("'anticlockwise' must be 'bool'")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.rotate_pitches(
                n_rotations=n_rotations,
                anticlockwise=anticlockwise,
            ))
        remove_repeated_time_signatures(dummy_container)
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    ### PRIVATE METHODS ###

    def _find_time_signatures(self):
        r"""Creates a list of all time signatures for all measures of
        ``contents``.
        """
        self._time_signatures = []
        leaves = abjad.select(self._current_window).leaves()
        duration = abjad.Duration(0)
        time_signature = abjad.inspect(
            leaves[0]).effective(abjad.TimeSignature)
        if not time_signature:
            time_signature = abjad.TimeSignature((4, 4))
        for leaf in leaves:
            if duration % time_signature.duration == 0:
                time_signature = abjad.inspect(
                    leaf).effective(abjad.TimeSignature)
                if time_signature:
                    duration = abjad.Duration(0)
                elif leaf is leaves[0]:
                    time_signature = abjad.TimeSignature((4, 4))
                else:
                    time_signature = self._time_signatures[-1]
                self._time_signatures.append(time_signature)
            duration += abjad.inspect(leaf).duration()
        self._time_signatures_durations = [timesig.duration for timesig
                                           in self._time_signatures]

    def _update_logical_ties(self):
        r'Updates the selection of logical ties of ``contents``.'
        self._logical_ties = abjad.select(self._current_window).logical_ties()

    def _get_pitch_list(self) -> list:
        r'Creates a list of all pitches in ``contents``.'
        pitches = []
        for logical_tie in abjad.select(self._current_window).logical_ties():
            leaf = logical_tie[0]
            if isinstance(leaf, abjad.Rest):
                pitches.append(None)
            elif isinstance(leaf, abjad.Note):
                pitches.append(leaf.written_pitch)
            elif isinstance(leaf, abjad.Chord):
                pitches.append(leaf.written_pitches)
        return pitches

    def _rewrite_pitches(self,
                         pitches: list,
                         ):
        r'Rewrites the pitches of all logical ties given a list of pitches.'
        index = 0
        dummy_container = abjad.Container(
            abjad.mutate(self._current_window[:]).copy()
        )
        logical_ties = abjad.select(dummy_container).logical_ties()
        for pitch, logical_tie in zip(pitches, logical_ties):
            for leaf in logical_tie:
                if not pitch:
                    new_leaf = abjad.Rest(leaf.written_duration)
                elif isinstance(pitch, abjad.PitchSegment):
                    new_leaf = abjad.Chord(pitch, leaf.written_duration)
                else:
                    new_leaf = abjad.Note(pitch, leaf.written_duration)
                for indicator in abjad.inspect(leaf).indicators():
                    if not pitch and isinstance(indicator, (abjad.Tie,
                                                            abjad.Articulation
                                                            )):
                        continue
                    if isinstance(indicator, abjad.TimeSignature):
                        if (not self._last_time_signature
                                or indicator != self._last_time_signature):
                            abjad.attach(indicator, new_leaf)
                            self._last_time_signature = indicator
                    else:
                        abjad.attach(indicator, new_leaf)
                selection = abjad.select(dummy_container).leaf(index)
                abjad.mutate(selection).replace(new_leaf)
                index += 1
        # output
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r'The ``abjad.Container`` to be shuffled.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: abjad.Container
                 ):
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child "
                            "class")
        self._contents = copy.deepcopy(contents)
        self._current_window = copy.deepcopy(contents)

    @property
    def output_single_measure(self) -> bool:
        r"""When ``True``, the output will be a single measure even if the
        contents of the shuffler are several measures.
        """
        return self._output_single_measure

    @output_single_measure.setter
    def output_single_measure(self,
                              output_single_measure: bool,
                              ):
        if not isinstance(output_single_measure, bool):
            raise TypeError("'output_single_measure' must be 'bool'")
        self._output_single_measure = output_single_measure

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the ``rewrite_meter`` mutation.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(self,
                              disable_rewrite_meter: bool,
                              ):
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def force_time_signatures(self) -> bool:
        r"""When ``True``, every call will output a selection with a time
        signature.
        """
        return self._force_time_signatures

    @force_time_signatures.setter
    def force_time_signatures(self,
                              force_time_signatures: bool,
                              ):
        if not isinstance(force_time_signatures, bool):
            raise TypeError("'force_time_signatures' must be 'bool'")
        self._force_time_signatures = force_time_signatures

    @property
    def omit_time_signatures(self) -> bool:
        r'When ``True``, the output will contain no time signatures.'
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(self,
                             omit_time_signatures: bool,
                             ):
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def current_window(self) -> abjad.Selection:
        r'Read-only property, returns the result of the last operation.'
        if self._omit_time_signatures:
            for leaf in abjad.select(self._current_window).leaves():
                for indicator in abjad.inspect(leaf).indicators():
                    if isinstance(indicator, abjad.TimeSignature):
                        abjad.detach(indicator, leaf)
        return copy.deepcopy(self._current_window)
